
{-# LANGUAGE ParallelListComp   #-}

module Main where

{-
   Notes:

    - Add 'Position' type ..
    - Verify random number selections
    - Ranges only used at initialize
-}

import System.Random
import Data.List (find, maximum, minimumBy)
import Data.Maybe

c1 = 1.4 --2.0
c2 = 1.4 --2.0
vMax = 2.0

maxNum = 99e99
ranges =  [(0.0, 10.0), (0.0, 10.0), (0.0, 10.0)] 

-- Particle data type representing one particle
data Particle = Particle {
                  pbest :: [Double],           -- particle's best solution
                  pbestFitness :: Double,
                  velocities :: [Double],      -- particle's velocities
                  present :: [Double],         -- particle's present solution
                  fitness :: Double            -- current fitness
                  } deriving (Show)

type FitnessCase = ([Double],[Double])

-- infinite list of infinite random lists (TODO: Remove fixed seed 42)
rList :: [[Double]]
rList = [randomRs ((0.0), 1.0) g | g <- gl (mkStdGen 42)]
   where gl gen = gen:(gl (fst $ split gen)) 

-- I wanted to use instance Ord but it required an Eq and I
-- didn't want to do == .. what does that mean on floats?

comparePs Particle {pbestFitness = p1}
              Particle {pbestFitness = p2} = compare p1 p2

--instance Ord Particle where
--  (<=) Particle {pbestFitness = p1}
--              Particle {pbestFitness = p2} = LT --compare p1 p2

cap :: Double -> Double
cap x | x < 0.0   = max x (-vMax)
      | otherwise = min x vMax    

-- Calculate the new velocities for a particle
calcVelocities :: Particle -> [Double] -> [Double] -> [Double]
calcVelocities p gbest rand = [ 0.73 * cap (v + c1 * r1 * (pb - pr) + c2 * r2 * (gb - pr))
                                    | v <- velocities p
                                    | pr <- present p
                                    | pb <- pbest p
                                    | gb <- gbest
                                    | r1 <- rand
                                    | r2 <- drop (length $ velocities p) rand]


-- Find the best global solution in the population
getBest :: [Particle] -> [Double]
getBest ps = pbest (minimumBy comparePs ps)
  where
    comparePs Particle {pbestFitness = p1}
              Particle {pbestFitness = p2} = compare p1 p2


-- Return the next generation

updateParticles :: [Particle] -> [FitnessCase] -> [Double] -> [Particle]
updateParticles population fCases rands =

          let psWithFitness = [ p { fitness = fitnessFunction (present p) fCases }
                                 | p <- population]

              psWithBest =  [if (fitness p < pbestFitness p)
                              then  p { pbest = (present p), pbestFitness = fitness p }
                              else  p  
                                | p <- psWithFitness ]

              gbest  = getBest psWithBest

              newVelocities =  map (\p -> calcVelocities p gbest rands) psWithBest  

              newParticles =  map (\(p,vs) -> p { present = [pr + v | pr <- present p
                                                                    | v <- vs ],
                                                  velocities = vs }) (zip psWithBest newVelocities)
          in newParticles 

-- Solve the problem using the ranges, fitness function and fitness cases
psoSolve :: [(Double, Double)] -> Int -> Int -> [FitnessCase] -> [Double]
psoSolve ranges numParticles numGenerations fCases = 
                  getBest $ foldl evolveGen (makeParticles numParticles) [1..numGenerations]

           where
            evolveGen :: [Particle] -> Int -> [Particle]
            evolveGen curPop genNum = updateParticles curPop fCases (head (drop (2 + genNum) rList))

hack :: [(Int,Int)]
hack = [(a,b) | a <- [1..3]
              | b <- [1..3]]

makeParticles :: Int ->  [Particle]
makeParticles count = makeParticles' count rList
  where
     makeParticles' :: Int -> [[Double]] -> [Particle]
     makeParticles' 0 _ = []
     makeParticles' count (r:rs) = 
          (makeParticle r) : (makeParticles' (count - 1) rs)

     makeParticle :: [Double] -> Particle
     makeParticle rs =
          Particle [] maxNum (randomVelocity rs) (randomPosition rs) maxNum
     
     randomVelocity :: [Double] -> [Double]
     randomVelocity rs = map calcRV (zip (take rangeLen rs) ranges)
     calcRV (randNo, range) = (fst range) + randNo * ((snd range) - (fst range))
 
     randomPosition :: [Double] -> [Double]
     randomPosition rs  = map calcRP (zip (take rangeLen (drop rangeLen rs)) ranges)
     calcRP (randNo, range) = (fst range) + randNo * ((snd range) - (fst range))

     rangeLen = length ranges
                
-------------------------------
--  Ad Hoc Fitness Function  --
-------------------------------

-- Fitness function takes the particles 'present' solution and returns a score
fitnessFunction :: [Double] -> [FitnessCase] -> Double
fitnessFunction xs fCases  =

 -- Here we assume that PSO will come up with 3 constants ..
 -- a0, a1, a2 such that
 --  y = a0 + a1*x + a2*x^2
 sum $ map (\fCase -> calcErr ((head (fst fCase)), (head (snd fCase))) xs) fCases

  where calcErr (x,out) (a0:a1:a2:a) = 
               (abs $ (a0*x*x + a1*x + a2) - out) -- smaller is better
        calcErr _ _ = maxNum

----------------------------
--  Ad Hoc Fitness Cases  --
----------------------------

fCases :: [FitnessCase]
fCases  =
   -- y = 3x^2 + 2x + 4
   -- [ (x0,y0),(x1,y1), (x2,y2) ................... (xn,yn) ]
      [  ([0],[4]) , ([1],[9]) , ([2],[20]),([3],[37]),([4],[60]),([5],[89]),([10],[324])]

--------------
--   Main   --  
--------------

main = do

         r <- return $ psoSolve ranges 20 100 fCases
         print r
         f <- return $ fitnessFunction r fCases
         print f

-- how to do randoms
-- setup population via one fitness run



