
{-# LANGUAGE ParallelListComp   #-}

module Main where

{-
   Notes:

    - Add 'Position' type ..
    - Verify random number selections
    - Ranges only used at initialize
-}

import System.Random
import Data.List (find, maximum, minimumBy, elemIndex, foldl1')
import Data.List.Split (chunksOf)
import Data.Maybe
import FCases

-----------------------------------------------
---------------  Parameters  ------------------
-----------------------------------------------

c1 = 2.0 --1.4 --2.0
c2 = 2.0 --1.4 --2.0
vMax = 30.0

particleCount = 40
generationCount = 500
inputCount =16
categoryCount = 7
randSeed = 123 -- 12345 -- 1234
maxNum = 99e99

-----------------------------------------------
-------------   Customization -----------------
-----------------------------------------------

data Category = Cat [Double]

ranges =  map (\i -> (-10.0,10.0)) [1..(inputCount * categoryCount)]

-- Fitness function takes the particles 'present' solution and returns a score
fitnessFunction :: [Double] -> [FitnessCase] -> Double
fitnessFunction answer fCases  =

 sum $ map (\fCase -> calcErr ((fst fCase), (head (snd fCase))) answer) fCases

  where
      calcErr :: ([Double], Double) -> [Double] -> Double 
      calcErr (inputs,out) answer =
         let allInputs = concat $ replicate categoryCount inputs
             products = map (\(inps,ans) -> (inps * ans)) (zip allInputs answer)
             --cGroups = chunksOf categoryCount products -- Thinking error!
             cGroups = chunksOf inputCount products

             -- Each group split at different position ..
             --cSplits = map (\(g,i) -> splitAt i g) cGroups
             sumOfGroups = map (\xs -> sum xs) cGroups
             cspl = splitAt (truncate out) sumOfGroups
             sumCat = last (fst cspl)
             noCats = init (fst cspl) ++ snd cspl
             sumNoCats = sum noCats
             numLess = sum $ map (\s -> if (s < sumCat) then 1 else 0) noCats
         in
             --foldl calcError 0.0 cSplits
             -- Sum all groups  but selected category group
             -- and multiply by selected category group value
                ---------------------------------------------------------------------------
             -- |  Cat 1  |  Cat 2  |  Cat 3  |  Cat 4  | Cat 5 |  ....
             -- | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | |

             --(last (fst cspl)) * ( sum (init (fst cspl)) + sum (snd cspl))
             --(last (fst cspl))
             -- (last (fst cspl)) + (if(sumNoCat < 2.0) then 1 else 0)
             numLess

calcHits :: [Double] -> [FitnessCase] -> (Int, Int)
calcHits gbest fCases =

    foldl (\(a,b) fCase -> ((a + (calcHit gbest fCase)), (b+1))) (0,0) fCases 

      where
        calcHit :: [Double] -> FitnessCase -> Int
        calcHit gbest (inputs,out:_) =
           let allInputs = concat $ replicate categoryCount inputs
               prod = map (\(f,b) -> f * b) (zip allInputs gbest)
               pspl = chunksOf inputCount prod
               cSums = map (\cs -> sum cs) pspl
               cAns = 1 + (fromJust (elemIndex (foldl1' min cSums ) cSums))
           in
               if (cAns == (truncate out)) then 1 else 0

-- Fitness cases defined externally ..

-----------------------------------------------
---------------- PSO  Engine ------------------
-----------------------------------------------


-- Particle data type representing one particle
data Particle = Particle {
                  pbest :: [Double],           -- particle's best solution
                  pbestFitness :: Double,
                  velocities :: [Double],      -- particle's velocities
                  present :: [Double],         -- particle's present solution
                  fitness :: Double            -- current fitness
                  }

instance Show Particle where
   show (Particle best bestF vs present fitness) =
      "\nP -> " ++ show present ++ "\n"

type FitnessCase = ([Double],[Double])

-- infinite list of infinite random lists (TODO: Remove fixed seed 42)
rList :: [[Double]]
rList = [randomRs ((0.0), 1.0) g | g <- gl (mkStdGen randSeed)]
   where gl gen = gen:(gl (fst $ split gen)) 

-- I wanted to use instance Ord but it required an Eq and I
-- didn't want to do == .. what does that mean on floats?

comparePs Particle {pbestFitness = p1}
              Particle {pbestFitness = p2} = compare p1 p2

--instance Ord Particle where
--  (<=) Particle {pbestFitness = p1}
--              Particle {pbestFitness = p2} = LT --compare p1 p2

cap :: Double -> Double
cap x | x < 0.0   = max x (-vMax)
      | otherwise = min x vMax    

-- Calculate the new velocities for a particle
calcVelocities :: Particle -> [Double] -> [Double] -> [Double]
calcVelocities p gbest rand = [ 0.73 * cap (v + c1 * r1 * (pb - pr) + c2 * r2 * (gb - pr))
                                    | v <- velocities p
                                    | pr <- present p
                                    | pb <- pbest p
                                    | gb <- gbest
                                    | r1 <- rand
                                    | r2 <- drop (length $ velocities p) rand]


-- Find the best global solution in the population
getBest :: [Particle] -> [Double]
getBest ps = pbest (minimumBy comparePs ps)
  where
    comparePs Particle {pbestFitness = p1}
              Particle {pbestFitness = p2} = compare p1 p2


-- Return the next generation

updateParticles :: [Particle] -> [FitnessCase] -> [Double] -> [Particle]
updateParticles population fCases rands =

          let psWithFitness = [ p { fitness = fitnessFunction (present p) fCases }
                                 | p <- population]

              psWithBest =  [if (fitness p < pbestFitness p)
                              then  p { pbest = (present p), pbestFitness = fitness p }
                              else  p  
                                | p <- psWithFitness ]

              gbest  = getBest psWithBest

              newVelocities =  map (\p -> calcVelocities p gbest rands) psWithBest  

              newParticles =  map (\(p,vs) -> p { present = [pr + v | pr <- present p
                                                                    | v <- vs ],
                                                  velocities = vs }) (zip psWithBest newVelocities)
          in newParticles 

-- Solve the problem using the ranges, fitness function and fitness cases
psoSolve :: [(Double, Double)] -> Int -> Int -> [FitnessCase] -> [Double]
psoSolve ranges numParticles numGenerations fCases = 
                  getBest $ foldl evolveGen (makeParticles numParticles) [1..numGenerations]

           where
            evolveGen :: [Particle] -> Int -> [Particle]
            evolveGen curPop genNum = updateParticles curPop fCases (head (drop (2 + genNum) rList))

hack :: [(Int,Int)]
hack = [(a,b) | a <- [1..3]
              | b <- [1..3]]

makeParticles :: Int ->  [Particle]
makeParticles count = makeParticles' count rList
  where
     makeParticles' :: Int -> [[Double]] -> [Particle]
     makeParticles' 0 _ = []
     makeParticles' count (r:rs) = 
          (makeParticle r) : (makeParticles' (count - 1) rs)

     makeParticle :: [Double] -> Particle
     makeParticle rs =
          Particle [] maxNum (randomVelocity rs) (randomPosition rs) maxNum
     
     randomVelocity :: [Double] -> [Double]
     randomVelocity rs = map calcRV (zip (take rangeLen rs) ranges)
     calcRV (randNo, range) = (fst range) + randNo * ((snd range) - (fst range))
 
     randomPosition :: [Double] -> [Double]
     randomPosition rs  = map calcRP (zip (take rangeLen (drop rangeLen rs)) ranges)
     calcRP (randNo, range) = (fst range) + randNo * ((snd range) - (fst range))

     rangeLen = length ranges
        
showHits :: [Particle] -> [FitnessCase] -> String
showHits ps fs =
  let hits = map (\p -> calcHits (present p)  fs) ps
  in
     concat $ map (\(a,b) -> "(" ++ show a ++ "," ++ show b ++ ")\n") hits

--------------
--   Main   --  
--------------

main = do
         --ps <- return $ makeParticles particleCount
         --putStrLn $ "All Hits -> " ++ show (map (\p -> (present p)) ps)
         --putStrLn $ "All Hits -> " ++ showHits ps fCases

         --a <- return $ calcHits (present (head ps)) fCases
         --putStrLn $ "Head Hits -> " ++ show a

         --a <- return $ calcHits (present (last ps)) fCases
         --putStrLn $ "Last Hits -> " ++ show a

         bestGuy <- return $ psoSolve ranges particleCount generationCount fCasesTrain
         --putStrLn $ "Best Guy -> " ++ show  bestGuy
         
         r <- return $ calcHits bestGuy fCasesTrain
         putStrLn $ "Train Hits -> " ++ show r

         r <- return $ calcHits bestGuy fCasesTest
         putStrLn $ "Test Hits -> "  ++ show r
         
         --f <- return $ fitnessFunction r fCases
         --print f

-- how to do randoms
-- setup population via one fitness run



